part of tecfy_core_client_package;

class TecfySyncService {
  static bool isSyncing = false;
  static get isOfflineMode {
    return !kIsWeb &&
        (ConnectionService.isOffline ||
            ConnectionService.activeConnection?.isCloud == true);
  }

  static List<String> _models;

  static Future<int> getUnpostedCount() async {
    var unPosted = await DatabaseService.rawQuery(
        'select count(*) from Entity where isPosted = 0');
    var deleted = await DatabaseService.rawQuery(
        'select count(*) from DeleteLog where isPosted = 0');
    if (unPosted.length == 0 && deleted.length == 0) return -1;

    return (unPosted.first.values.first + deleted.first.values.first);
  }

  static Future<List<dynamic>> getUnpostedRecords() async {
    var unPosted = await DatabaseService.rawQuery(
        "select id,entity,timeStamp,message,'Update' as type from Entity where isPosted = 0");
    var deleted = await DatabaseService.rawQuery(
        "select id,entity,timeStamp,message,'Delete' as type from DeleteLog where isPosted = 0");
    var list = List<dynamic>();
    list.addAll(unPosted);
    list.addAll(deleted);
    return list;
  }

  static Future<bool> undoDelete(String id) async {
    var result = await DatabaseService.delete('DeleteLog',
        where: 'id = ?', whereArgs: [id]);
    return result > 0;
  }

  static Future<bool> pushNow() async {
    try {
      var isLoggin = await AccountService.isLogin(null);
      if (kIsWeb || ConnectionService.isOffline || isLoggin.data != true) {
        return true;
      }
      print('++++++++++++++++++++++++++++++++ Start Sync');

      //====================================================
      //======  delete local
      //====================================================
      print('============= delete local');
      var lastSyncDeletedInt = ConfigService.getValueInt('lastSyncDeleted');
      DateTime lastSyncDeleted = DateTime.now().toUtc();
      if (lastSyncDeletedInt != null && lastSyncDeletedInt > 0)
        lastSyncDeleted =
            DateTime.fromMillisecondsSinceEpoch(lastSyncDeletedInt);
      lastSyncDeleted = lastSyncDeleted.add(Duration(hours: -5));
      var serverDeleted = await ApiService.httpGetDynamic(
          'sync/getDeleted/${intl.DateFormat("yyyy-MM-ddTHH:mm:ss").format(lastSyncDeleted)}');
      if (serverDeleted.data != null && serverDeleted.data.length > 0) {
        var idsToDelete =
            serverDeleted.data.map<String>((e) => e['id'].toString()).toList();
        print(idsToDelete);
        await DatabaseService.deleteEntityBulk(idsToDelete);

        (serverDeleted.data as List<dynamic>)
            .sort((a, b) => b['timeStamp'].compareTo(a['timeStamp']));

        ConfigService.setValueInt(
            'lastSyncDeleted',
            BaseModel.parseUTC(serverDeleted.data.first['timeStamp'])
                .millisecondsSinceEpoch);
      }
      //====================================================
      //======  delete from server
      //====================================================
      print('============= delete from server');
      var localDeleted =
          (await DatabaseService.query('DeleteLog', where: 'isPosted = 0'))
              .toList();

      if (localDeleted.length > 0) {
        var sendServerDeleteResult = await ApiService.httpPost(
            'sync/bulkDelete',
            localDeleted.map((e) {
              return {'id': e['id'], 'entity': e['entity']};
            }).toList(),
            true);

        print('+++++++++++++++++++++++++++++++++++ sendServerDeleteResult');
        print(sendServerDeleteResult.data);
        if (sendServerDeleteResult.success &&
            sendServerDeleteResult.data != null &&
            sendServerDeleteResult.data.length > 0) {
          print(sendServerDeleteResult.data);
          var deleteResult = sendServerDeleteResult.data as List<dynamic>;
          await DatabaseService.exeBatch((b) async {
            for (var i in deleteResult) {
              if (i['success']) {
                b.delete('DeleteLog', where: 'id = ?', whereArgs: [i['id']]);
              } else {
                b.update(
                    'DeleteLog', {'message': 'Item related to other items'},
                    where: 'id = ?', whereArgs: [i['id']]);
              }
            }
            b.commit();
            return true;
          });
        }
      }
      //====================================================
      //======  Post to server
      //====================================================
      var unPosted =
          (await DatabaseService.query('Entity', where: 'isPosted = 0'))
              .toList();
      print('SSSSSSSSSSSSS-------------- unPosted to sync ${unPosted.length}');
      if (unPosted.length == 0) return true;

      //===========
      //== Post un-posted items
      //===========
      await _getModels();

      for (var model in _models) {
        model = EntityService.lowerFirst(model);
        print('=============== start $model');
        for (var item in unPosted.where((e) => e['entity'] == model)) {
          print(item['id']);
          var result = await ApiService.httpPostEx(
              'entity/addupdate/$model',
              jsonEncode(jsonDecode(item['body']),
                      toEncodable: EntityService.customEncode)
                  .replaceAll('\"', '\''));
          if (result.success) {
            await DatabaseService.update(
                'Entity', {'isPosted': 1, 'message': ''},
                where: 'id = ?', whereArgs: [item['id']]);
          } else {
            await DatabaseService.update('Entity', {'message': result.message},
                where: 'id = ?', whereArgs: [item['id']]);
            print(result.toString());
          }
        }
      }
      return true;
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXXXXXXXXX Error Sync');
      print(err);
      return false;
    }
  }

  static _getModels() async {
    if (_models != null) return;
    var modelsResult = await ApiService.httpGetDynamic('Sync/getModels');
    if (!modelsResult.success) return false;
    _models = modelsResult.data;
  }

  static Future<bool> syncNow() async {
    if (kIsWeb) {
      return true;
    }
    await pushNow();

    if (LocalConfigService.operationMode == OperationMode.online) return true;

    await _getModels();

    var lastSyncEntityInt = ConfigService.getValueInt('lastSyncEntity');
    DateTime lastSyncEntity = DateTime.now().toUtc();
    if (lastSyncEntityInt != null && lastSyncEntityInt > 0)
      lastSyncEntity = DateTime.fromMillisecondsSinceEpoch(lastSyncEntityInt);
    lastSyncEntity = lastSyncEntity.add(Duration(hours: -5));

    for (var model in _models) {
     var result = await ApiService.httpGetDynamic('entity/get/$model?count=1000');
     var oldData = await DatabaseService.query(model,where:'isPosted= 0');
     await DatabaseService.delete(model,where:'isPosted= 1');
     var oldIds= oldData.map((d)=>d['id']).toList();
     await DatabaseService.addEntityAll(model, result.data.where((r)=> !oldIds.contains(r['id'])).toList(),isPosted:true);
    }
    //===========
    //== Load Item category
    //===========
    // var categoryResult =
    //     await ApiService.httpGetDynamic('pos/GetItemCategorySecondry');

    // if (categoryResult.success) {
    //   var items = (categoryResult.data as List<dynamic>)
    //       .map<ItemCategory>((e) => ItemCategory.fromJson(e, null))
    //       .toList();

    //   await DatabaseService.delete('ItemCategory');

    //   await DatabaseService.exeBatch((batch) async {
    //     for (var item in items) {
    //       batch.insert('ItemCategory', item.toJson());
    //     }
    //     return true;
    //   });
    //   print(
    //       '+++++++++++++++++++ ItemCategory count ${(await DatabaseService.rawQuery('select count(*) from ItemCategory'))}');
    // } else {
    //   print('XXXX Error ItemCategory ${categoryResult.message} ');
    // }

    return true;
  }
}
